namespace eval tktocanvas {

proc setRotation {point angle} {

    variable rotAngle
    variable rotPoint
    
    set rotPoint $point
    set rotAngle $angle
}

proc setScaleFactors {sx sy} {

    variable scaleX
    variable scaleY

    set scaleX $sx
    set scaleY $sy
}

proc setOffsets {x y} {

    variable offsetX
    variable offsetY

    set offsetX $x
    set offsetY $y
}

proc ScaleText {textcoords fontString} {

	set xy [DoScaling [lindex $textcoords 0] [lindex $textcoords 1]]
	set font [ScaleFont $fontString]

    return [list $xy $font]
}

proc ScaleFont {fontString} {
	
    lappend tmpFont [font actual $fontString -family]
	set size [font actual $fontString -size]

	if {$Size < 0} {
		set size [expr {-$size}]
	}

	lappend tmpFont [expr {int($Size * $::tktocanvas::scaleX)}]
	lappend tmpFont [font actual $fontString -slant]
	lappend tmpFont [font actual $fontString -weight]
	
	return $tmpFont
}

proc ScaleCoords {twoPts x y {x1 0} {y1 0}} {

    set coords {}

	foreach {x y} [DoScaling $x $y] {
		lappend coords $x
		lappend coords $y
	}

    if {$twoPts} {
        foreach {x y} [DoScaling $x1 $y1] {
            lappend coords $x
            lappend coords $y
        }
    }

    return $coords
}

proc ScaleCoordsList {twoPts coords} {
    
    return [ScaleCoords $twoPts [lindex $coords 0] [lindex $coords 1] \
            [lindex $coords 2] [lindex $coords 3]]
}

proc ScaleImage {imagecoords imagewidth imageheight} {

    set coords [ScaleCoords 0 [lindex $imagecoords 0] [lindex $imagecoords 1]]
    set iWidth [expr {int($imageheight * $::tktocanvas::scaleX)}]
    set iHeight [expr {int($imagewidth * $::tktocanvas::scaleY)}]

    return [list $coords $iWidth $iHeight]
}

proc DoScaling {x y} {

    set scaledX 0
    set scaledY 0

    if {$::tktocanvas::scaleX != 1} {
        set scaledX [expr {$::tktocanvas::scaleX - 1}]
        set scaledX [expr {int($x * $scaledX)}]
    }

    if {$::tktocanvas::scaleY != 1} {
        set scaledY [expr {$::tktocanvas::scaleY - 1}]
        set scaledY [expr {int($y * $scaledY)}]
    }

    set x [expr {$x + $scaledX}]
    set y [expr {$y + $scaledY}]

    return [list $x $y]
}

#############################################################################
# Procedure rotateXY --
#
#       Rotates point (x, y) relative to datum point (x0, y0) by angle.
#
# Arguments:
#       x, y    - Point to be rotated.
#       rotPt   - Datum point.
#       angle   - Rotation angle.
# Results:
#       Returns point which is rotated relative to (x0, y0) by angle.

proc rotateXY {x y} {

    set angle [expr {$::tktocanvas::rotAngle * atan(1) * 4 / 180.0}]

    set x [expr $x - [lindex $::tktocanvas::rotPoint 0]]
    set y [expr $y - [lindex $::tktocanvas::rotPoint 1]]

    set newX [expr {int($x * cos($angle) - $y * sin($angle))}]
    set newY [expr {int($x * sin($angle) + $y * cos($angle))}]

    incr newX [lindex $::tktocanvas::rotPoint 0]
    incr newY [lindex $::tktocanvas::rotPoint 1]

    return [list $newX $newY]
}

proc rotateXYList {coords} {

    return [rotateXY [lindex $coords 0] [lindex $coords 1]]
}

proc RotateImage {imgCoords imgWidth imgHeight} {
   
    foreach {x y} [rotateXYList $imgCoords] break
    
    return [list $x $y]
    
    switch $::tktocanvas::rotAngle {
        90 { 
            foreach {x y} [rotateXYList $imgCoords] {
                #set y [expr {$y + ($::tktocanvas::scaleX * $imgWidth)}]
            }
        }
        180 {
            foreach {x y} [rotateXYList $imgCoords] {
                #set x [expr {$x + ($::tktocanvas::scaleX * $imgWidth)}]
                #set y [expr {$y - ($::tktocanvas::scaleY * $imgHeight)}]
            }
        }
        270 {
            foreach {x y} [rotateXYList $imgCoords] {
                #set x [expr {$x + ($::tktocanvas::scaleY * $imgHeight)}]
            }
        }
    }

}

proc ExchangePoints {xName yName x1Name y1Name} {
    
    upvar $xName x $yName y $x1Name x1 $y1Name y1

    switch $::tktocanvas::rotAngle {
        90 {
            set tmp $x1
            set x1 $x
            set x $tmp
        }
        180 {
            set tmp $x1
            set tmp2 $y1
            set x1 $x
            set y1 $y
            set x $tmp
            set y $tmp2
        }
        270 {
            set tmp $y1
            set y1 $y
            set y $tmp
        }
    }
}

proc TransformImage {coords width height} {
   
    if {($::tktocanvas::scaleX != 1) || ($::tktocanvas::scaleY != 1)} {
        set coords [ScaleCoordsList 0 $coords]
    }

    if {$::tktocanvas::rotAngle != 0} {
        set coords [RotateImage $coords $width $height]
    }
    
    lset coords 0 [expr {[lindex $coords 0] + $::tktocanvas::offsetX}]
    lset coords 1 [expr {[lindex $coords 1] + $::tktocanvas::offsetY}]

    return $coords
}

proc TransformDictPoints {dictName {x x} {y y} {x1 x1} {y1 y1}} {

    upvar $dictName dict

    if {($::tktocanvas::scaleX != 1) || ($::tktocanvas::scaleY != 1)} {
        foreach {xp yp x1p y1p} [ScaleCoords 1 [dict get $dict -$x] \
                [dict get $dict -$y] [dict get $dict -$x1] \
                [dict get $dict -$y1]] {

                dict set dict -$x $xp
                dict set dict -$y $yp
                dict set dict -$x1 $x1p
                dict set dict -$y1 $y1p
        }
    }
    
    if {$::tktocanvas::rotAngle != 0} {
        foreach {xp yp x1p y1p} "[rotateXY [dict get $dict -$x] \
                 [dict get $dict -$y]] [rotateXY [dict get $dict -$x1] \
                 [dict get $dict -$y1]]" {

            ExchangePoints xp yp x1p y1p

            dict set dict -$x $xp
            dict set dict -$y $yp
            dict set dict -$x1 $x1p
            dict set dict -$y1 $y1p
        }
    }

    dict set dict -$x [expr {round([dict get $dict -$x] \
            + $::tktocanvas::offsetX)}] 
    dict set dict -$y [expr {round([dict get $dict -$y] \
            + $::tktocanvas::offsetY)}] 
    dict set dict -$x1 [expr {round([dict get $dict -$x1] \
            + $::tktocanvas::offsetX)}] 
    dict set dict -$y1 [expr {round([dict get $dict -$y1] \
            + $::tktocanvas::offsetY)}] 
}

proc TransformPoints {xName yName x1Name y1Name} {

    upvar $xName x $yName y $x1Name x1 $y1Name y1

    if {($::tktocanvas::scaleX != 1) || ($::tktocanvas::scaleY != 1)} {
        foreach {xp yp x1p y1p} [ScaleCoords 1 $x $y $x1 $y1] {

                set x $xp
                set y $yp
                set x1 $x1p
                set y1 $y1p
        }
    }
    
    if {$::tktocanvas::rotAngle != 0} {
        foreach {xp yp x1p y1p} "[rotateXY $x $y] [rotateXY $x1 $y1]" {

            ExchangePoints xp yp x1p y1p

            set x $xp
            set y $yp
            set x1 $x1p
            set y1 $y1p
        }
    }

    set x [expr {round($x + $::tktocanvas::offsetX)}]
    set y [expr {round($y + $::tktocanvas::offsetY)}]
    set x1 [expr {round($x1 + $::tktocanvas::offsetX)}]
    set y1 [expr {round($y1 + $::tktocanvas::offsetY)}]
}

proc TransformPointsLst {ptsList} {

    set res {}

    foreach {x y} $ptsList {
        foreach {x y} [ScaleCoords 0 $x $y] break
        
        if {$::tktocanvas::rotAngle != 0} {
            foreach {x y} [rotateXY $x $y] break
        }
        
        set x [expr {round($x + $::tktocanvas::offsetX)}]
        set y [expr {round($y + $::tktocanvas::offsetY)}]

        lappend res $x $y
    }

    return $res
}

proc TransformPointsList {pointList} {

    set listLength [llength $pointList]

    if {$listLength != 4} {
        TransformPointsLst $pointList
    } else {
        foreach {x y x1 y1} $pointList break
       
        if {($::tktocanvas::scaleX != 1) || ($::tktocanvas::scaleY != 1)} {
            foreach {xp yp x1p y1p} [ScaleCoords $listLength $x $y $x1 $y1] {

                    set x $xp
                    set y $yp
                    set x1 $x1p
                    set y1 $y1p
            }
        }
        
        if {$::tktocanvas::rotAngle != 0} {
            foreach {xp yp x1p y1p} "[rotateXY $x $y] [rotateXY $x1 $y1]" {

                ExchangePoints xp yp x1p y1p

                set x $xp
                set y $yp
                set x1 $x1p
                set y1 $y1p
            }
        }

        set x [expr {round($x + $::tktocanvas::offsetX)}]
        set y [expr {round($y + $::tktocanvas::offsetY)}]
        set x1 [expr {round($x1 + $::tktocanvas::offsetX)}]
        set y1 [expr {round($y1 + $::tktocanvas::offsetY)}]
        
        return [list $x $y $x1 $y1]
    }
}

} ;# end of namespace tktocanvas
